#!/usr/bin/env python
""" Create an input file for the mobcal program from a computational chemistry
log file.

Isotope masses correspond to the most abundent form of each atom, pulled from
Mathematica 9.0's ElementData function. See
http://reference.wolfram.com/language/note/ElementDataSourceInformation.html
for references to the source of the isotope masses.

Written by Adam B. Birkholz (abirkhol@amgen.com) and may not be modified
or redistributed without permission

"""

import os
import argparse

_ISOTOPES = {
    'H' : 1, 'He' : 4, 'Li' : 7, 'Be' : 9, 'B' : 11, 'C' : 12, 'N' : 14,
    'O' : 16, 'F' : 19, 'Ne' : 20, 'Na' : 23, 'Mg' : 24, 'Al' : 27, 'Si' : 28,
    'P' : 31, 'S' : 32, 'Cl' : 35, 'Ar' : 40, 'K' : 39, 'Ca' : 40, 'Sc' : 45,
    'Ti' : 48, 'V' : 51, 'Cr' : 52, 'Mn' : 55, 'Fe' : 56, 'Co' : 59, 'Ni' : 58,
    'Cu' : 63, 'Zn' : 64, 'Ga' : 69, 'Ge' : 74, 'As' : 75, 'Se' : 80,
    'Br' : 79, 'Kr' : 84, 'Rb' : 85, 'Sr' : 88, 'Y' : 89, 'Zr' : 90, 'Nb' : 93,
    'Mo' : 98, 'Ru' : 102, 'Rh' : 103, 'Pd' : 106, 'Ag' : 107, 'Cd' : 114,
    'In' : 115, 'Sn' : 120, 'Sb' : 121, 'Te' : 130, 'I' : 127, 'Xe' : 132,
    'Cs' : 133, 'Ba' : 138, 'La' : 139, 'Ce' : 140, 'Pr' : 141, 'Nd' : 142,
    'Sm' : 152, 'Eu' : 153, 'Gd' : 158, 'Tb' : 159, 'Dy' : 164, 'Ho' : 165,
    'Er' : 166, 'Tm' : 169, 'Yb' : 174, 'Lu' : 175, 'Hf' : 180, 'Ta' : 181,
    'W' : 184, 'Re' : 187, 'Os' : 192, 'Ir' : 193, 'Pt' : 195, 'Au' : 197,
    'Hg' : 202, 'Tl' : 205, 'Pb' : 208, 'Bi' : 209, 'Th' : 232, 'U' : 238
}

def to_mass(sym, iso):
    """ Return the approximate atomic mass for something with atomic symbol
    sym. Use masses in iso first
    """

    if sym in iso:
        return iso[sym]
    else:
        return _ISOTOPES[sym]

def clean_symbol(sym):
    """ Return the atomic regardless of the case of the input sym """
    return sym[0].upper() + sym[1:].lower()

def parse_gaussian(filename):
    """Read the geometry, atoms, and ESP charges from a pop=MK g09 calculation.

    The atoms and charges are picked up from the ESP charges section. When
    multiple such sections are found in the log file, only the last one is
    returned.

    Similarly, the coordinates are read from the ESP "Atomic Center" lists, and
    only the last such set are found.

    If the format of the gaussian log file changes in the future, this may need
    to be updated
    """

    coords = []
    incharge = False
    base = os.path.basename(os.path.splitext(filename)[0])

    with open(filename, 'r') as file:
        for line in file:
            if line[1:13] == "ESP charges:":
                incharge = True
                charges = []
                atoms = []
            elif incharge:
                if line[1:19] == "Sum of ESP charges":
                    incharge = False
                split = line.split()
                if len(split) == 3:
                    charges.append(float(split[2]))
                    atoms.append(split[1])
            elif line[7:20] == "Atomic Center":
                split = line.split()
                coords.append([float(c) for c in split[5:8]])

    if len(coords) > len(atoms):
        coords = coords[-len(atoms):]

    return base, coords, atoms, charges

def parse_gamess(filename):
    """Read the geometry, atoms, and ESP charges from a gamess calculation.

    Coordinates are picked up from the "coordinates of all atoms" section, and
    only the last such set are returned. Similarly, atoms and charges are
    picked up from the "net charges" section, and only the last set are
    returned.

    If the format of the gamess log file chages in the future, this may need to
    be updated
    """

    incoord = False
    incharge = False
    base = os.path.basename(os.path.splitext(filename)[0])

    with open(filename, 'r') as file:
        for line in file:
            if line[1:13] == "NET CHARGES:":
                incharge = True
                charges = []
                atoms = []
            elif incharge:
                if not line.strip():
                    incharge = False
                elif line[1:5] != "ATOM" and line[1:5] != "----":
                    split = line.split()
                    charges.append(float(split[1]))
                    atoms.append(clean_symbol(split[0]))
            elif line[1:25] == "COORDINATES OF ALL ATOMS":
                incoord = True
                coords = []
            elif incoord:
                if not line.split():
                    incoord = False
                elif line[3:7] != "ATOM" and line[3:7] != "----":
                    split = line.split()
                    coords.append([float(c) for c in split[2:5]])

    return base, coords, atoms, charges

def read_masses(filename):
    """Read the atomic masses from the file"""

    masses = []
    with open(filename, 'r') as file:
        for line in file:
            split = line.split()
            if len(split) > 1:
                raise(Exception("line in mass file has more than one item"))
            elif len(split) == 1:
                masses.append(int(round(float(split[0]), 0)))
            else:
                # assume blank line is at the end, do nothing
                continue

    return masses

def read_isotopes(filename):
    """Generate a dictionary mapping atomic numbers to custom, read in isotopes
    """

    isotopes = {}

    with open(filename, 'r') as file:
        for line in file:
            split = line.split()
            if len(split) != 2:
                raise(Exception("line in isotope file has incorrect number of items"))
            else:
                atms = clean_symbol(split[0])
                if atms in isotopes:
                    raise(Exception("Too many definitions for {0} in isotope file".format(atms)))
                else:
                    isotopes[atms] = int(round(float(split[1]), 0))

    return isotopes

def prep_mfj(base, coords, atoms, charges):
    """Print the mfj file to filename"""

    natoms = len(coords)
    to_write = []

    to_write.append("Input={0}\n".format(base))
    to_write.append("1\n")
    to_write.append("{0}\n".format(natoms))
    to_write.append("ang\n")
    to_write.append("calc\n")
    to_write.append("1.0000\n")
    for i,xc in enumerate(coords):
        to_write.append(
            " {x[0]:13.5f}{x[1]:13.5f}{x[2]:13.5f}{a:>5d}{c:14.6f}\n".format(
                x=xc, a=atoms[i], c=charges[i]
            )
        )

    return ''.join(to_write)

if __name__ == "__main__":
    argparser = argparse.ArgumentParser(
        description="Generate mfj file from an ESP fit log file. Use massfile" +
        " or isotopefile to read in custom atomic masses",
        argument_default=argparse.SUPPRESS
    )
    argparser.add_argument(
        'logfile', metavar='LOGFILE',
        help="log file containing the results of ESP fit calculation"
    )
    argparser.add_argument(
        'mfjfile', metavar='MFJFILE', default=None, nargs='?',
        help='Output file, omit to print to stdout'
    )
    argparser.add_argument(
        '-m', '--massfile', metavar='FILE', default=None,
        help="Override atomic mass assignment for all atoms. Text file with" +
        " natom masses on separate lines"
    )
    argparser.add_argument(
        '-i', '--isotopefile', metavar='FILE', default=None,
        help="Override atomic mass assignment by atom symbol. Text file with" +
        " symbol and mass on individual lines (e.g. C 13 or H 2)"
    )
    argparser.add_argument(
        '-t', '--type', default='gaussian', type=str.lower,
        choices=['gaussian','gamess'], help="Program used to produce log" +
        " file. Default is gaussian"
    )

    args = vars(argparser.parse_args())
    logfile = args['logfile']
    mfjfile = args['mfjfile']
    massfile = args['massfile']
    isofile = args['isotopefile']
    logtype = args['type']

    if logtype == 'gaussian':
        base, crd, atm, chg = parse_gaussian(logfile)
    elif logtype == 'gamess':
        base, crd, atm, chg = parse_gamess(logfile)

    if massfile is not None:
        if isofile is not None:
            raise(Exception("isofile and massfile should not both be specified"))
        atmmass = read_masses(massfile)
        assert len(atmmass) == len(crd)
    else:
        if isofile is not None:
            isotopes = read_isotopes(isofile)
        else:
            isotopes = {}
        atmmass = [to_mass(a, isotopes) for a in atm]

    mfjstring = prep_mfj(base, crd, atmmass, chg)

    if mfjfile is not None:
        with open(mfjfile, 'w') as file:
            file.write(mfjstring)
    else:
        print(mfjstring)
